"""Implementation of RbaParameters, RbaFunction and RbaAggregate classes.

Peter Schubert, CCB, HHU Duesseldorf, December 2022
"""

import os
import numpy as np
import pandas as pd
from xml.etree.ElementTree import parse, ElementTree, Element, SubElement, indent

from ..utils.rba_utils import record_generator, get_function_params


def extract_params(record):
    """Extract parameters from a record.

    A single record consists of comma separated key-value pairs.
    Example: 'key1=val1, key2=val2, ...' is converted to
    {key1: val1, key2: val2, ...}

    :param record: key '=' value pairs separated by ","
    :type record: str
    :returns: key-values pairs extracted from record
    :rtype: dict
    """
    params = {}
    if type(record) is str:
        for kv_pair in record_generator(record, sep=','):
            if '=' in kv_pair:
                k, v = kv_pair.split('=')
                params[k.strip()] = float(v.strip())
    return params


class RbaParameters:

    def __init__(self):
        self.functions = {}
        self.aggregates = {}
        self.f_name_zero = 'zero'
        self.f_name_spontaneous = 'default_spontaneous'

    def import_xml(self, model_dir):

        file_name = os.path.join(model_dir, 'parameters.xml')
        if os.path.exists(file_name) is True:
            root = parse(file_name).getroot()
            assert root.tag == 'RBAParameters'
            self.functions = RbaFunction.import_xml(root.find('listOfFunctions'))
            self.aggregates = RbaAggregate.import_xml(root.find('listOfAggregates'))
        else:
            print(f'{file_name} not found!')

    def from_df(self, m_dict):
        if 'functions' in m_dict:
            self.functions = RbaFunction.from_df(m_dict['functions'])
        else:
            print(f'functions not imported!')
        if 'aggregates' in m_dict:
            self.aggregates = RbaAggregate.from_df(m_dict['aggregates'])
        else:
            print(f'aggregates not imported!')

    def create_parameter(self, f_name_base, constant, func, agg):
        """Create a parameter (function or aggregate)

        Create constant, general function or aggregate.

        :param str f_name_base: base name of parameter name
        :param constant: constant parameter
        :type constant: int/float or nan
        :param func: function definition with required keywords
        :type func: str or nan
        :param agg: comma separated list if function names
        :type agg: str or nan
        :return: parameter name or None
        :rtype: str or Name
        """
        p_name = None
        if np.isfinite(constant):
            p_name = f'{f_name_base}_const'
            f_params = {'type': 'constant', 'variable': 'growth_rate', 'params': {'CONSTANT': constant}}
            self.add_function(p_name, f_params)
        elif type(func) is str:
            p_name = f'{f_name_base}_func'
            f_params = get_function_params(func)
            self.add_function(p_name, f_params)
        elif type(agg) is str:
            p_name = f'{f_name_base}_agg'
            f_names = [item.strip() for item in agg.split(',')]
            self.add_aggregate(p_name, f_names)
        return p_name

    def from_xba(self, rba_params):
        """Add functions from RBA specific parameters.

        Mainly adds complex functions from rba_params['functions'] used in aggregates
        Few constant functions 'zero' and 'default_spontaneous' for spontaneous reactions
        Most functions and aggregates are created while processing other RBA components
        Add default efficiency functions for most frequent kcat values

        :param rba_params: RBA model specific parametrization
        :type rba_params: dict of pandas DataFrames
        """
        for f_name, row in rba_params['functions'].iterrows():
            const_value = row['constant']
            if np.isfinite(const_value):
                self.functions[f_name] = RbaFunction(f_name, f_type='constant', f_params={'CONSTANT': const_value})
            else:
                assert (type(row['function']) is str)
                f_params = get_function_params(row['function'])
                self.functions[f_name] = RbaFunction(f_name, f_type=f_params['type'], f_params=f_params['params'],
                                                     f_variable=f_params.get('variable', 'growth_rate'))

        # ensure that 'zero' function and default kcat for spontaneous reactions are included
        if self.f_name_zero not in self.functions:
            f_name = self.f_name_zero
            self.functions[f_name] = RbaFunction(f_name, f_type='constant', f_params={'CONSTANT': 0.0})

        if self.f_name_spontaneous not in self.functions:
            f_name = self.f_name_spontaneous
            self.functions[f_name] = RbaFunction(f_name, f_type='constant', f_params={'CONSTANT': 100.0 * 3600.0})

    def add_function(self, f_name, f_params):
        self.functions[f_name] = RbaFunction(f_name, f_type=f_params['type'], f_params=f_params['params'],
                                             f_variable=f_params.get('variable', 'growth_rate'))

    def add_aggregate(self, agg_name, f_names):
        self.aggregates[agg_name] = RbaAggregate(agg_name, agg_type='multiplication', f_names=f_names)

    def get_values(self, params):
        """Get function and aggregate values
        :param params: variables with their values
        :type params: dict (keys: variable id / str, val: value / float)
        :return: parameter values values determined based on params
        :rtype: dict (key: function/aggregate id / str, val value / float)
        """
        for f in self.functions.values():
            f.set_value(params)
        for agg in self.aggregates.values():
            agg.set_value(self.functions)

        parameter_values = {}
        for fid, f in self.functions.items():
            parameter_values[fid] = f.value
        for agg_id, agg in self.aggregates.items():
            parameter_values[agg_id] = agg.value
        return parameter_values

    def export_xml(self, model_dir):

        file_name = os.path.join(model_dir, 'parameters.xml')
        root = Element('RBAParameters')

        functions = SubElement(root, 'listOfFunctions')
        for item in self.functions.values():
            functions.append(item.export_xml())

        aggregates = SubElement(root, 'listOfAggregates')
        for item in self.aggregates.values():
            aggregates.append(item.export_xml())

        tree = ElementTree(root)
        indent(tree)
        tree.write(file_name)

    def get_value_info(self, value):
        if value in self.functions:
            value_info = self.functions[value].get_value_info()
        else:
            value_info = self.aggregates[value].get_value_info()
        return value_info

    def to_df(self):
        df_f = pd.DataFrame([item.to_dict() for item in self.functions.values()])
        df_f.set_index('function', inplace=True)
        df_a = pd.DataFrame([item.to_dict() for item in self.aggregates.values()])
        df_a.set_index('aggregate', inplace=True)
        return {'functions': df_f, 'aggregates': df_a}

    def validate(self, component_ids):
        valid = True
        missing = self.ref_functions(self.aggregates).difference(component_ids['functions'])
        if len(missing) > 0:
            print('functions used in aggregates not defined:', missing)
            valid = False
        return valid

    def ref_functions(self, ref_parameters):
        # add functions used in aggregates
        refs = set()
        for param in ref_parameters:
            if param in self.aggregates:
                refs |= set(self.aggregates[param].functions)
        return refs


class RbaFunction:

    def __init__(self, fid, f_type='', f_params=None, f_variable='growth_rate'):
        self.id = fid
        self.type = f_type
        self.variable = f_variable
        self.parameters = f_params if type(f_params) is dict else {}
        self.value = 0.0

    @staticmethod
    def import_xml(functions):

        data = {}
        for function in functions.findall('function'):
            fid = function.attrib['id']
            rba_function = RbaFunction(fid)
            rba_function.type = function.attrib['type']
            rba_function.variable = function.attrib['variable']
            parameters = function.find('listOfParameters')
            if parameters is not None:
                for parameter in parameters.findall('parameter'):
                    pid = parameter.attrib['id']
                    value = float(parameter.attrib['value'])
                    rba_function.parameters[pid] = value
            data[fid] = rba_function
        return data

    @staticmethod
    def from_df(df):
        data = {}
        for fid, row in df.iterrows():
            rba_function = RbaFunction(fid)
            rba_function.type = row['type']
            rba_function.variable = row['variable']
            rba_function.parameters = extract_params(row['parameters'])
            data[fid] = rba_function
        return data

    def export_xml(self):
        attribs = {'id': self.id, 'type': self.type, 'variable': self.variable}
        function = Element('function', attribs)

        if len(self.parameters) > 0:
            parameters = SubElement(function, 'listOfParameters')
            for parameter, value in self.parameters.items():
                SubElement(parameters, 'parameter', {'id': parameter, 'value': str(round(value, 8))})
        return function

    def set_value(self, params):
        """Set function value based parameters provided.

        :param params: variables with their values
        :type params: dict (keys: variable id / str, val: value / float)
        """
        value = 0.0
        param_val = params.get(self.variable, 0.0)
        if self.type == 'indicator':
            value = 1 if self.parameters["X_MIN"] <= param_val <= self.parameters["X_MAX"] else 0
        else:
            if 'X_MIN' in self.parameters:
                param_val = max(param_val, self.parameters['X_MIN'])
            if 'X_MAX' in self.parameters:
                param_val = min(param_val, self.parameters['X_MAX'])

            if self.type == 'constant':
                value = self.parameters["CONSTANT"]
            elif self.type == 'linear':
                value = self.parameters["LINEAR_CONSTANT"] + param_val * self.parameters["LINEAR_COEF"]
            elif self.type == 'michaelisMenten':
                value = self.parameters["kmax"] * param_val / (self.parameters["Km"] + param_val)
            elif self.type == 'exponential':
                value = np.exp(self.parameters["RATE"] * param_val)

            if 'Y_MIN' in self.parameters:
                value = max(value, self.parameters['Y_MIN'])
            if 'Y_MAX' in self.parameters:
                value = min(value, self.parameters['Y_MAX'])
        self.value = round(value, 8)

    def get_value_info(self):
        """Convert Function Configuration into a string representation

        :return: value information
        :rtype: str
        """
        value_info = f'{self.type}: '
        if self.type == 'constant':
            value_info += f'{self.parameters["CONSTANT"]}'
        elif self.type == 'linear':
            value_info += (f'{self.parameters["LINEAR_CONSTANT"]} + ' +
                           f'({self.parameters["LINEAR_COEF"]} * {self.variable})' +
                           f' [{self.parameters["X_MIN"]}, {self.parameters["X_MAX"]}]' +
                           f' -> [{self.parameters["Y_MIN"]}, {self.parameters["Y_MAX"]}]')
        elif self.type == 'michaelisMenten':
            value_info += (f'{self.parameters["kmax"]} * {self.variable}/' +
                           f'({self.parameters["Km"]} + {self.variable})' +
                           f' -> [{self.parameters.get("Y_MIN", "0")}, inf]')
        elif self.type == 'indicator':
            value_info += f'{self.variable} in [{self.parameters["X_MIN"]}, {self.parameters["X_MAX"]}]'
        elif self.type == 'exponential':
            value_info += f'exp({self.parameters["RATE"]} * {self.variable})'
        return value_info

    def to_dict(self):
        parameters = ', '.join([f'{pid}={value}' for pid, value in self.parameters.items()])
        return {'function': self.id, 'type': self.type, 'variable': self.variable, 'parameters': parameters}


class RbaAggregate:

    def __init__(self, aggid, agg_type='multiplication', f_names=None):
        self.id = aggid
        self.type = agg_type
        self.functions = f_names if type(f_names) is list else []
        self.value = 1.0

    @staticmethod
    def import_xml(aggregates):

        data = {}
        for aggregate in aggregates.findall('aggregate'):
            aid = aggregate.attrib['id']
            rba_aggregate = RbaAggregate(aid)
            rba_aggregate.type = aggregate.attrib['type']
            function_refs = aggregate.find('listOfFunctionReferences')
            for function_ref in function_refs.findall('functionReference'):
                rba_aggregate.functions.append(function_ref.attrib['function'])
            data[aid] = rba_aggregate
        return data

    @staticmethod
    def from_df(df):
        data = {}
        for aid, row in df.iterrows():
            rba_aggregate = RbaAggregate(aid)
            rba_aggregate.type = row['type']
            rba_aggregate.functions = [item.strip() for item in row['functions'].split(',')]
            data[aid] = rba_aggregate
        return data

    def export_xml(self):
        attribs = {'id': self.id, 'type': self.type}
        aggregate = Element('aggregate', attribs)

        function_refs = SubElement(aggregate, 'listOfFunctionReferences')
        for function_ref in self.functions:
            SubElement(function_refs, 'functionReference', {'function': function_ref})
        return aggregate

    def set_value(self, functions):
        value = 1.0
        for fid in self.functions:
            value *= functions[fid].value
        self.value = round(value, 8)

    def get_value_info(self):
        return f'aggregate: {", ".join(self.functions)}'

    def to_dict(self):
        return {'aggregate': self.id, 'type': self.type, 'functions': ', '.join(self.functions)}



# -*- coding: utf-8 -*-
import os
import re
from setuptools import setup, find_packages


setup_kwargs = {}

with open('README.md') as f:
    setup_kwargs['long_description'] = f.read()

# version from file
with open(os.path.join('f2xba', '_version.py')) as f:
    mo = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                   f.read(), re.MULTILINE)
    if mo:
        setup_kwargs['version'] = mo.group(1)

setup(
    name='f2xba',
    description='Support creation of XBA models from FBA models',
    author='Peter Schubert',
    author_email='peter.schubert@hhu.de',
    url='https://gitlab.cs.uni-duesseldorf.de/schubert/f2xba',
    project_urls={
        "Source Code": 'https://gitlab.cs.uni-duesseldorf.de/schubert/f2xba',
        "Bug Tracker": 'https://gitlab.cs.uni-duesseldorf.de/schubert/f2xba/-/issues'},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    license='GPLv3',
    long_description_content_type='text/markdown',
    packages=find_packages(exclude='docs'),
    install_requires=['pandas>=1.4.0',
                      'numpy>=0.21.0',
                      'scipy>=1.11.0',
                      'requests>=2.26.0',
                      'matplotlib>=3.6.3',
                      'sbmlxdf>=0.2.7'],
    python_requires=">=3.8",
    keywords=['modeling', 'SBML', 'FBA', 'RBA', 'GBA',
              'bioinformatics'],
    **setup_kwargs
)



"""Implementation of FbcObjective class.

Peter Schubert, HHU Duesseldorf, June 2022
"""

import sbmlxdf

from .sbml_sbase import SbmlSBase


class FbcObjective(SbmlSBase):

    def __init__(self, s_fbc_objective):
        """Instantiate FbcObjective instance with data from s_fbc_objective

        s_reaction is pandas Series with Series.name attribute set reaction id and
        several mandatory and optional attributes based on SBML specifications.

        - mandatory attributes:
            - Series.name: str - objective id
            - 'type': str - 'maximize' or 'minimize'
            - 'active': bool
            - 'fluxObjectives': str - with species references

        - optional attributes:
            - 'name': str - handled in parent class
            - 'sboterm': str - handled in parent class
            - 'metaid': str - handled in parent class
            - 'miriamAnnotation': str - handled in parent class
            - 'notes': str - handled in parent class

        :param s_fbc_objective: Fbc Objective data from SBML import
        :type s_fbc_objective: pandas Series
        """
        super().__init__(s_fbc_objective)
        self.direction = s_fbc_objective['type']
        self.active = s_fbc_objective['active']
        if 'coefficients' in s_fbc_objective:
            self.coefficients = s_fbc_objective['coefficients']
        else:
            self.coefficients = {}
            for reac_ref in sbmlxdf.record_generator(s_fbc_objective['fluxObjectives']):
                params = sbmlxdf.extract_params(reac_ref)
                self.coefficients[params['reac']] = float(params['coef'])

    def to_dict(self):
        data = super().to_dict()
        data['type'] = self.direction
        data['active'] = self.active
        data['fluxObjectives'] = '; '.join([f'reac={rid}, coef={coef}' for rid, coef in self.coefficients.items()])
        return data

    def modify_attribute(self, attribute, value):
        """modify attribute value.

        :param attribute: attribute name
        :type attribute: str
        :param value: value to be configured
        :type value: str
        """
        if hasattr(self, attribute):
            setattr(self, attribute, value)
        else:
            print(f'unknown FbcObjective attribute {attribute}')


"""Implementation of BiocycRNA class.

Peter Schubert, CCB, HHU Duesseldorf, November 2022
"""

import re
import xml.etree.ElementTree

from f2xba.utils.biocyc_utils import get_child_text, get_sub_obj_ids


class BiocycRNA:

    def __init__(self, biocyc_id):
        self.id = biocyc_id
        self.name = ''
        self.synonyms = ''
        self.gene = None
        self.complexes = []
        self.gene_composition = {}

    @staticmethod
    def get_rnas(file_name):
        """Retrieve RNA data from BioCyc export.

        """
        tree = xml.etree.ElementTree.parse(file_name)
        root = tree.getroot()

        data = {}
        for el in root.findall('RNA'):
            biocyc_id = el.get('ID').split(':')[1]
            bc_rna = BiocycRNA(biocyc_id)
            bc_rna.name = get_child_text(el, 'common-name')
            bc_rna.synonyms = re.sub(r'\|', ',', get_child_text(el, 'synonym'))
            genes = get_sub_obj_ids(el, 'gene', 'Gene')
            if len(genes) == 1:
                bc_rna.gene = genes[0]
            bc_rna.complexes = get_sub_obj_ids(el, 'component-of', '*')
            data[biocyc_id] = bc_rna
        return data



"""Implementation of TdMReactionData class.

holding thermodynamics data for a reaction
based on pyTFA thermo data

Peter Schubert, HHU Duesseldorf, Octobert 2023
"""
# import re


class TdReactionData:

    def __init__(self, rid, reversible, kind):
        """Instantiate thermodynamics reaction data.

        based on pyTFA thermo_data.thermodb

        Note: for strings we convert the type from numpy.str_ to str

        :param str rid: reaction id
        :param bool reversible: reversible status from genome scale metabolic model
        :param str kind: reaction type, e.g. 'metabolic' or 'transport', from original GEM
        """
        self.id = rid
        self.reversible = reversible
        self.kind = kind
        self.drg0_tr = None
        self.drg0_tr_error = None
        self.add_td_constraints = False

    def modify_attribute(self, attribute, value):
        """modify attribute value.

        :param str attribute: attribute name
        :param value: value to be configured
        :type value: str, float, bool
        """
        if hasattr(self, attribute):
            setattr(self, attribute, value)
        else:
            print(f'unknown TD metabolite attribute {attribute}')



"""Implementation of Protein class.

Holding protein information for genes in the model

Note: for handling of wildcard EC numbers, see f2xba package

Peter Schubert, HHU Duesseldorf, June 2023
"""


class Protein:

    def __init__(self, uniprot, locus, compartment=''):
        self.id = uniprot.id
        self.name = uniprot.protein_name
        self.gene_name = uniprot.gene_name
        self.mw = uniprot.mass
        self.length = uniprot.length
        self.ec_numbers = sorted(uniprot.ec_numbers)
        self.location = uniprot.location
        self.locus = locus
        self.compartment = compartment
        self.linked_sids = set()
        self.up_cofactors = uniprot.cofactors
        self.cofactors = {}
        self.aa_sequence = uniprot.aa_sequence
        self.has_signal_peptide = True if type(uniprot.signal_peptide) is str else False

    def set_compartment(self, compartment):
        self.compartment = compartment

    @property
    def cid(self):
        return self.compartment.split('-')[0]

    def link_sid(self, sid):
        self.linked_sids.add(sid)

    def modify_attribute(self, attribute, value):
        """modify attribute value.

        :param attribute: attribute name
        :type attribute: str
        :param value: value to be configured
        :type value: str
        """
        setattr(self, attribute, value)



"""Implementation of RbaMedium class.

Peter Schubert, CCB, HHU Duesseldorf, December 2022
"""

import os
import pandas as pd

MAX_MEDIUM_CONC = 10.0


class RbaMedium:

    def __init__(self):
        self.concentrations = {}

    def import_xml(self, model_dir):
        # actually an import from tsv file
        file_name = os.path.join(model_dir, 'medium.tsv')
        if os.path.exists(model_dir) is True:
            df = pd.read_table(file_name, usecols=['Metabolite', 'Concentration'], index_col='Metabolite')
            self.concentrations = df.to_dict()['Concentration']
        else:
            print(f'{file_name} not found!')

    def from_df(self, m_dict):
        if 'medium' in m_dict:
            self.concentrations = m_dict['medium'].to_dict()['Concentration']
        else:
            print(f'medium not imported!')

    def from_xba(self, general_params, xba_model):
        """Configure Medium based xba_model reactions.

        Medium defines the nutrient environment for the model simulation.
        In FBA based simulations, exchange reactions import metabolites into the model.
        Import of a specific metabolite is allowed when the corresponding
        exchange reaction is unblocked in the updake direction (usually a negative lower flux bound).

        RBA modelling does not use exchange reactions, instead if defines positive
        metabolite concentrations for the selected environment.
        These concentrations are configured in the RBA medium component.

        Therefore, potential nutrients are defined based on available exchange reactions
        in the metabolic model (FBA model).
        An initial medium is configured based on the flux bounds of exchange reactions.

        Metabolites in the RBA do not contain compartment postfixes.
            i.e. metabolite M_glc__D instead of species id M_glc__D_e

        In RBA, species in the external compartment (those created via exchange reactions in FBA)
        are configured with boundaryCondition set to True, i.e. their concentration is not
        affected during simulation
        Usually ,using Outer Membrane Transporters, external species are transported
        from external compartment to periplasm. These OM transporters will have an associated
        protein cost linear with the total import rate.

        In E. coli metabolite concentration in external medium and periplasm are close to identical.
        RBA then implement Michaelis Menten kinetics for importing species from periplasm to cytosol,
        using as species concentration the respective concentration of the medium metabolite.

        Maximum medium concentration from rba_params['general']['medium_max_conc']['value']

        :param general_params: general RBA parameters loaded from file
        :type general_params: dict
        :param xba_model: xba model based on genome scale metabolic model
        :type xba_model: Class XbaModel
        """
        medium_max_conc = general_params.get('medium_max_conc', MAX_MEDIUM_CONC)

        nutrients = 0
        for rid, r in xba_model.reactions.items():
            if r.kind == 'exchange':
                if len(r.reactants) == 1:
                    sid = list(r.reactants)[0]
                    flux_bnd = xba_model.parameters[r.fbc_lower_bound].value
                else:
                    sid = list(r.products)[0]
                    flux_bnd = xba_model.parameters[r.fbc_upper_bound].value
                mid, cid = sid.rsplit('_', 1)
                assert (cid == xba_model.external_compartment)
                if abs(flux_bnd) > 0.0:
                    nutrients += 1
                self.concentrations[mid] = min(abs(flux_bnd), medium_max_conc)
        print(f'{len(self.concentrations):4d} medium metabolites ({nutrients} > 0.0 mmol/l)')

    def export_xml(self, model_dir):
        # actually an export to tsv file
        file_name = os.path.join(model_dir, 'medium.tsv')
        df = self.to_df()['medium']
        df.to_csv(file_name, sep='\t')

    def to_df(self):
        df = pd.DataFrame(self.concentrations.values(), index=list(self.concentrations),
                          columns=['Concentration'])
        df.index.name = 'Metabolite'
        return {'medium': df}



"""Implementation of RbaDensities and RbaDensity classes.

Peter Schubert, CCB, HHU Duesseldorf, December 2022
"""

import os
import numpy as np
import pandas as pd
from xml.etree.ElementTree import parse, ElementTree, Element, SubElement, indent

from .rba_target_value import RbaTargetValue
from ..utils.rba_utils import extract_params


class RbaDensities:

    def __init__(self):
        self.densities = {}

    def import_xml(self, model_dir):
        file_name = os.path.join(model_dir, 'density.xml')
        if os.path.exists(model_dir) is True:
            root = parse(file_name).getroot()
            assert root.tag == 'RBADensity'
            self.densities = RbaDensity.import_xml(root.find('listOfTargetDensities'))
        else:
            print(f'{file_name} not found!')

    def from_df_old(self, m_dict):
        if 'densities' in m_dict:
            self.densities = RbaDensity.from_df(m_dict['densities'])
        else:
            print(f'densities not imported!')

    def from_xba(self, rba_params, parameters):
        """Configure Density Constraints based on RBA sepecific parameters.

        used: rba_params['compartments']
        functions and aggregates are added to Parameters

        :param rba_params: RBA model specific parametrization
        :type rba_params: dict of pandas DataFrames
        :param parameters: RBA model parameters
        :type parameters: Class RbaParameters
        """
        for c_name, row in rba_params['compartments'].iterrows():
            value_type = row.get('density_constraint_value_type', 'upperBound')
            constant = row.get('density_constraint_constant', np.nan)
            func = row.get('density_constraint_function', np.nan)
            agg = row.get('density_constraint_aggregate', np.nan)
            p_name = parameters.create_parameter(f'{c_name}_density', constant, func, agg)
            if p_name:
                target_value = RbaTargetValue.get_target_value(value_type, p_name)
                self.densities[c_name] = RbaDensity(c_name, target_value)
        print(f'{len(self.densities):4d} density constraints added')

    def export_xml(self, model_dir):

        file_name = os.path.join(model_dir, 'density.xml')
        root = Element('RBADensity')

        target_densities = SubElement(root, 'listOfTargetDensities')
        for item in self.densities.values():
            target_densities.append(item.export_xml())

        tree = ElementTree(root)
        indent(tree)
        tree.write(file_name)

    def to_df(self):
        df = pd.DataFrame([item.to_dict() for item in self.densities.values()])
        df.set_index('compartment', inplace=True)
        return {'densities': df}

    def validate(self, component_ids):
        valid = True
        missing_components = self.ref_parameters().difference(component_ids['functions'])\
                                                  .difference(component_ids['aggregates'])
        if len(missing_components) > 0:
            print('function/aggregates used in densities not defined:', missing_components)
            valid = False
        return valid

    def ref_parameters(self):
        refs = set()
        for d in self.densities.values():
            refs |= set(d.target_value.to_dict().values())
        return refs


class RbaDensity:

    def __init__(self, cid, target_value=None):
        self.id = cid
        self.target_value = target_value
        self.constr_id = None

    @staticmethod
    def import_xml(target_densities):
        data = {}
        for target_density in target_densities.findall('targetDensity'):
            cid = target_density.attrib['compartment']
            rba_density = RbaDensity(cid)
            rba_density.target_value = RbaTargetValue.from_dict(target_density.attrib)
            data[cid] = rba_density
        return data

    @staticmethod
    def from_df(df):
        data = {}
        for cid, row in df.iterrows():
            rba_density = RbaDensity(cid)
            target_value = extract_params(row['targetValue'])
            rba_density.target_value = RbaTargetValue.from_dict(target_value)
            data[cid] = rba_density
        return data

    def export_xml(self):
        attribs = self.target_value.to_dict() | {'compartment': self.id}
        return Element('targetDensity', attribs)

    def to_dict(self):
        return {'compartment': self.id, 'targetValue': self.target_value.to_str()}



"""Prefixes.py

Variable and Constraint prefixes used across different models

Peter Schubert, CCB, HHU Duesseldorf, Mai 2024
"""

# general prefixes
R_ = 'R_'         # reaction prefix
M_ = 'M_'         # prefix for mass balance constraint (species)
V_ = 'V_'         # general variable prefix, if not reaction
C_ = 'C_'         # general constraint prefix

# variable prefixes - ECM related
V_PC_ = f'{V_}PC_'                           # protein concentration
V_PC_total = f'{V_}PC_total'                 # total protein (modeled)
V_PSLACK_ = f'{V_}PSLACK_'                   # positive protein concentration slack
V_NSLACK_ = f'{V_}NSLACK_'                   # negative protein concentration slack

# constraint prefixes - ECM related
C_prot_ = f'{C_}prot_'            # protein concentration
C_prot_pool = f'{C_}prot_pool'  # total protein pool

# variable prefixes - RBA related
R_PROD_ = f'{R_}PROD_'    # production machinery reaction
R_DEGR_ = f'{R_}DEGR_'    # degradation machinery reaction
V_PMC_ = f'{V_}PMC_'      # protein mass fraction
V_EC_ = f'{V_}EC_'        # enzyme concentration
V_TMMC_ = f'{V_}TMMC_'    # target concentration for macromolecule
V_SLACK_ = f'{V_}SLACK_'  # slack on compartment density

V_TSMC = f'{V_}TSMC'     # target concentration for small molecules
V_TCD = f'{V_}TCD'       # target compartment density

# constraint prefixes - RBA related
MM_ = 'MM_'           # Macromolecule mass balance
C_PMC_ = f'{C_}PMC_'     # process machine capacity
C_EF_ = f'{C_}EF_'       # forward enzyme capacity
C_ER_ = f'{C_}ER_'       # reverse enzyme capacity
C_D_ = f'{C_}D_'         # compartment density

# variable prefixes - TD modeling related
V_DRG_ = f'{V_}DRG_'    # ∆rG', transformed Gibbs Free Energy of Reaction
V_DRG0_ = f'{V_}DRG0_'  # ∆rGo', standard transformed Gibbs Free Energy of Reaction
V_FU_ = f'{V_}FU_'      # forward reaction use variable (binary)
V_RU_ = f'{V_}RU_'      # reverse reaction use variable (binary)
V_LC_ = f'{V_}LC_'      # log concentration variable for metabolite
V_PS_ = f'{V_}PS_'      # positive slack on log concentration
V_NS_ = f'{V_}NS_'      # negative slack on log concentration
V_LNG_ = f'{V_}LNG_'    # thermodynamic displacement

V_RHS_FC = f'{V_}RHS_FC'  # righthand side flux coupling
V_RHS_GC = f'{V_}RHS_GC'  # righthand side energy coupling

# constraint prefixes - TD modeling related
C_DRG_ = f'{C_}DRG_'    # ∆rG', transformed Gibbs Free Energy of Reaction
C_SU_ = f'{C_}SU_'      # simultaneous use
C_GFC_ = f'{C_}GFC_'    # ∆rG' forward coupling
C_GRC_ = f'{C_}GRC_'    # ∆rG' reverse coupling
C_FFC_ = f'{C_}FFC_'    # flux forward coupling
C_FRC_ = f'{C_}FRC_'    # flux reverse coupling
C_DC_ = f'{C_}DC_'      # ∆rG' displacement constraint



"""Implementation of TdReactantData class.

Biochemical reactions deal with reactants, e.g. ATP
Reactants may be composed of several 'chemical' species, e.g. ATP4-, HATP3-, H2ATP2-
TD reactant has one-to-one relation with a model species
TD reactant is also linked to a single TD species with thermodynamics data

holding thermodynamic data for a TD reactant, like ATP
based on pyTFA thermo data

Peter Schubert, HHU Duesseldorf, Octobert 2023
"""


class TdReactantData:

    def __init__(self, sid, td_rdata):
        """Instantiate thermodynamic reactant data.

        contains reactant TD data for reactant at specific pH and ionic strength
        based on pyTFA thermo data file

        :param str sid: model species id
        :param dict td_rdata: thermodynamics data for TD reactant
        """
        self.id = sid
        self.td_sid = td_rdata['td_sid']
        self.ph = td_rdata['ph']
        self.ionic_str = td_rdata['ionic_str']
        self.dfg0_tr = td_rdata.get('dfg0_tr')
        self.avg_h_atoms_tr = None
        self.avg_charge_tr = None

    def modify_attribute(self, attribute, value):
        """modify attribute value.

        :param str attribute: attribute name
        :param value: value to be configured
        :type value: str, float, int
        """
        if hasattr(self, attribute):
            setattr(self, attribute, value)
        else:
            print(f'unknown TD reactant attribute {attribute}')



"""Implementation of BiocycData class.

Peter Schubert, CCB, HHU Duesseldorf, November 2022
"""

import os
import pandas as pd
import urllib.parse
import urllib.request

from .biocyc_gene import BiocycGene
from .biocyc_protein import BiocycProtein
from .biocyc_rna import BiocycRNA


class BiocycData:

    def __init__(self, biocyc_dir, org_prefix='ecoli', api_url='https://websvc.biocyc.org/xmlquery'):
        """Instantiate BiocycModel Instance

        For a given organism, query Biocyc on-line database for specific components in
         suitable detail level. Subsequently, extract relevant information.
        Use already downloaded Biocyc exports in case they exist in biocyc_dir.

        :param biocyc_dir: directory name, where downloads of Biocyc are stored
        :type biocyc_dir: str
        :param org_prefix:
        :param api_url:
        """

        self.prefix_lc = org_prefix.lower()
        self.prefix_uc = org_prefix.upper() + ':'
        self.url = api_url
        self.biocyc_dir = biocyc_dir

        # Exports with required level of detail.
        self.biocyc_data = {'Gene': ['genes', 'full'],
                            'Protein': ['proteins', 'low'],
                            'RNA': ['RNAs', 'low']}

        if self.is_complete_biocyc_data() is False:
            self.retrieve_biocyc_data()

        self.genes = BiocycGene.get_genes(self.biocyc_data_fname('Gene'))
        self.locus2gene = {gene.locus: bc_id for bc_id, gene in self.genes.items()}
        self.proteins = BiocycProtein.get_proteins(self.biocyc_data_fname('Protein'))
        self.rnas = BiocycRNA.get_rnas(self.biocyc_data_fname('RNA'))

        # set gene locus on direct gene product and rnas
        for protein in self.proteins.values():
            if type(protein.gene) is str and protein.gene in self.genes:
                protein.gene_composition = {self.genes[protein.gene].locus: 1.0}
        for rna in self.rnas.values():
            if type(rna.gene) is str and rna.gene in self.genes:
                rna.gene_composition = {self.genes[rna.gene].locus: 1.0}

    @staticmethod
    def add_composition(tot_composition, composition, p_stoic):
        for item, stoic in composition.items():
            if item not in tot_composition:
                tot_composition[item] = 0
            tot_composition[item] += stoic * p_stoic
        return tot_composition

    def add_simple_proteins(self, add_proteins):
        """Add simple proteins (i.e. direct gene products).

        Proteins have a unique Protein id, e.g. 'GATC-MONOMER'
          and parameters, supplied in a dict
        Protein instance is created, configured and added to dict of
          existing proteins in Biocyc model.
        Gene needs to exist.
        Gene is identified by gene locus, e.g. 'b2092'
        Gene record is updated with new protein id

        :param add_proteins: proteins with configuration data
        :type add_proteins: dict of dict
        :return: number of added proteins
        :rtype: int
        """
        n_updates = 0
        for pid, data in add_proteins.items():
            if 'locus' in data:
                gid = self.locus2gene[data['locus']]
                p = BiocycProtein(pid)
                p.gene = gid
                p.gene_composition = {data['locus']: 1.0}
                self.proteins[pid] = p
                self.modify_proteins({pid: data})
                if pid not in self.genes[gid].proteins:
                    self.genes[gid].proteins.append(pid)
                n_updates += 1
        return n_updates

    def modify_proteins(self, modify_proteins):
        """Modify configuration of proteins.

        Proteins have a unique Protein id, e.g. 'CPLX0-231'

        :param modify_proteins: proteins with configuration data
        :type modify_proteins: dict of dict
        :return: number of added proteins
        :rtype: int
        """
        n_updates = 0
        for pid, data in modify_proteins.items():
            if pid in self.proteins:
                p = self.proteins[pid]
                for key, val in data.items():
                    if hasattr(p, key):
                        setattr(p, key, val)
                n_updates += 1
        return n_updates

    def get_gene_composition(self, protein_id):
        """Iteratively retrieve for a complex the protein/rna gene loci with stoichiometry

        :param protein_id: ecocyc protein id
        :type protein_id: str
        :return: dictionary with gene composition
        :rtype: dict (key: gene locus, str; value: stoic, float)
        """
        gene_composition = {}
        p = self.proteins[protein_id]

        if len(p.gene_composition) > 0:
            # in case a protein is a direct gene product, its composition contains the gene reference.
            # create new dict with gene composition to avoide side effects
            gene_composition = {gene: stoic for gene, stoic in p.gene_composition.items()}
        else:
            # retrieve composition information for a protein complex (iteratively)
            for p_part_id, p_stoic in p.protein_parts.items():
                composition = self.get_gene_composition(p_part_id)
                gene_composition = self.add_composition(gene_composition, composition, p_stoic)
            # add rna gene composition
            for rna_id, p_stoic in p.rna_parts.items():
                composition = self.rnas[rna_id].gene_composition
                gene_composition = self.add_composition(gene_composition, composition, p_stoic)
        return gene_composition

    def get_cofactor_composition_old(self, protein_id):
        """Iteratively retrieve cofactor composition with stoichiometry

        Note: Cofactor composition in Ecocyc protein export is not complete.
              use Uniprot Cofactors instead

        :param protein_id: ecocyc protein id
        :type protein_id: str
        :return: dictionary with cofactor composition
        :rtype: dict (key: ecocyc compound id, str; value: stoic, float)
        """
        p = self.proteins[protein_id]
        cofactor_composition = p.compound_parts
        for p_part_id, p_stoic in p.protein_parts.items():
            composition = self.get_cofactor_composition_old(p_part_id)
            if len(composition) > 0:
                cofactor_composition = self.add_composition(cofactor_composition, composition, p_stoic)
        return cofactor_composition

    def set_enzyme_composition(self):
        """Retrieve enzyme gene and cofactor composition on protein data

        Cofactor composition in Biocyc protein export is not complete.
        Instead, we use the cofactor composition information stored in Uniprot
        """
        for protein_id, p in self.proteins.items():
            if len(p.protein_parts) > 0:
                p.gene_composition = self.get_gene_composition(protein_id)
                # p.cofactor_composition = self.get_cofactor_composition(protein_id)

    def biocyc_data_fname(self, component):
        class_name, detail = self.biocyc_data[component]
        return os.path.join(self.biocyc_dir, f'biocyc_{class_name}_{detail}.xml')

    def is_complete_biocyc_data(self):
        exports_available = True
        for component in self.biocyc_data:
            file_name = self.biocyc_data_fname(component)
            if os.path.exists(file_name) is False:
                exports_available = False
                print(f'{file_name} does not exist.')
        return exports_available

    def retrieve_biocyc_data(self):
        """Retrieve data from Biocyc site using REST API

        using BioVelo query, see https://biocyc.org/web-services.shtml

        """
        for component in self.biocyc_data:
            file_name = self.biocyc_data_fname(component)
            if os.path.exists(file_name) is False:
                class_name, detail = self.biocyc_data[component]
                print('retrieve ' + class_name + ' from Biocyc at level ' + detail)
                base_url = urllib.parse.urlparse(self.url)
                biovelo_qry = f'[x: x<-{self.prefix_lc}^^{class_name}]'
                query = urllib.parse.quote(biovelo_qry) + '&detail=' + detail
                split_url = base_url._replace(query=query)
                url = urllib.parse.urlunparse(split_url)
                req = urllib.request.Request(url)

                with urllib.request.urlopen(req) as response, open(file_name, 'w') as file:
                    file.write(response.read().decode('utf-8'))
                print(f'{file_name} from biocyc retrieved')

    def get_protein_components(self, protein_id):
        """Determine components of a given protein/enzyme.

        recursively parse through proteins and extract compounds
        and loci for proteins, rnas. All with stoichiometry
        """

        components = {'proteins': {}, 'rnas': {}}   # 'compounds': {}}
        gene = self.proteins[protein_id].gene
        if gene is not None:
            locus = self.genes[gene].locus
            components['proteins'] = {locus: 1.0}
        else:
            for rna_part in self.proteins[protein_id].rna_parts:
                rna, stoic_str = rna_part.split(':')
                gene = self.rnas[rna].gene
                locus = self.genes[gene].locus
                stoic = int(stoic_str)
                if locus not in components['rnas']:
                    components['rnas'][locus] = stoic
                else:
                    components['rnas'][locus] += stoic
            for protein_part in self.proteins[protein_id].protein_parts:
                protein, stoic_str = protein_part.split(':')
                stoic = int(stoic_str)
                sub_components = self.get_protein_components(protein)
                # update components:
                for part_type in sub_components:
                    for component in sub_components[part_type]:
                        if component not in components[part_type]:
                            components[part_type][component] = stoic * sub_components[part_type][component]
                        else:
                            components[part_type][component] += stoic * sub_components[part_type][component]
        return components

    def export_enzyme_composition(self, fname):
        """Export enzyme composition to Excel document.

        after initializeing BiocyData(biocyc_dir) and
        configuration set_enzyme_composition

        :param fname: file name of export file
        :type fname: str
        """
        enz_comp = {}
        for enz_id, enz in self.proteins.items():
            if len(enz.gene_composition) > 0 and len(enz.enzrxns) > 0:
                gene_comp = '; '.join([f'gene={gene}, stoic={stoic}'
                                       for gene, stoic in enz.gene_composition.items()])
                enz_comp[enz_id] = [enz.name, enz.synonyms, gene_comp]

        df_enz_comp = pd.DataFrame(enz_comp.values(), index=list(enz_comp), columns=['name', 'synonyms', 'genes'])
        df_enz_comp.index.name = 'biocyc_id'

        with pd.ExcelWriter(fname) as writer:
            df_enz_comp.to_excel(writer, sheet_name='e_coli')
            print(f'{len(df_enz_comp)} enzyme compositions written to {fname}')



"""Tools supporting EC, RBA and TFA model optimization using CobraPy

Peter Schubert, 18.04.2024 (HHU, CCB)
"""
from .fba_opt import FbaOptimization
from .fba_results import FbaResults
from .ecm_opt import EcmOptimization
from .ecm_results import EcmResults
from .rba_opt import RbaOptimization
from .rba_results import RbaResults
from .fba_opt import FbaOptimization
from .gecko_fit_kcats import GeckoFitKcats

__all__ = ['FbaOptimization', 'FbaResults',
           'EcmOptimization', 'EcmResults',
           'RbaOptimization', 'RbaResults'
           ,GeckoFitKcats]



"""
f2xba package
=============

Package supporting creation XBA  models.

Peter Schubert, CCB, HHU-Duesseldorf, June 2023
"""

from .xba_model import *
from .ec_model import EcModel
from .rba_model import RbaModel
from .tfa_model import TfaModel
from .utils.mapping_utils import update_master_kcats

from . import utils
from .biocyc import BiocycData
from .uniprot import UniprotData
from .ncbi import NcbiData

from .optimization.fba_opt import FbaOptimization
from .optimization.ecm_opt import EcmOptimization
from .optimization.rba_opt import RbaOptimization

from .optimization.fba_results import FbaResults
from .optimization.ecm_results import EcmResults
from .optimization.rba_results import RbaResults

from .optimization.gecko_fit_kcats import GeckoFitKcats


__all__ = []
__all__ += xba_model.__all__
__all__ += ec_model.__all__
__all__ += biocyc.__all__
__all__ += uniprot.__all__
__all__ += ncbi.__all__
__all__ += rba_model.__all__
__all__ += tfa_model.__all__



"""Implementation InitialAssignments required for RBA optimization.

Supporting both CobraPy or GurobiPy models

Class IaFunction
Class IaTargetRid
Class InitialAssignments

Peter Schubert, HHU Duesseldorf, January 2024
"""


import re
import numpy as np
from collections import defaultdict

import sbmlxdf
import f2xba.prefixes as pf


class IaFunction:

    def __init__(self, symbol, math_str):
        self.symbol = symbol
        self.math = math_str
        self.expanded_math = None
        self.is_medium_dependent = False
        self.is_growth_rate_dependent = False

    def set_expanded_math(self, func_defs, unit_ids):
        """Expand the initial assignment math string, so it can be evaluated

        :param func_defs: funtion definitions of the model
        :param unit_ids: unit ids that will be removed from the math string
        """
        math_str = sbmlxdf.misc.mathml2numpy(self.math)

        # remove any unit ids
        for uid in unit_ids:
            math_str = re.sub(r'\b' + uid + r'\b', '', math_str)

        # expand the math string using function definitions
        for fid, data in func_defs.items():
            m = re.search(r'\b' + fid + r'\(([^)]*)\)', math_str)
            while m:
                values = [val.strip() for val in m.group(1).split(',')]
                f_inline = data['definition']
                f_params = data['parameters']
                for param, value in zip(f_params, values):
                    f_inline = re.sub(r'\b' + param + r'\b', value, f_inline)
                math_str = math_str.replace(m.group(0), f'({f_inline})')
                m = re.search(r'\b' + fid + r'\(([^)]*)\)', math_str)
        self.expanded_math = math_str

    def get_value(self, local_env):
        return eval(self.expanded_math, {}, local_env)


class IaTargetRid:

    def __init__(self, rid):
        self.rid = rid
        self.flux_bounds = {}
        self.reactants = {}
        self.products = {}
        self.is_medium_dependent = False
        self.is_growth_rate_dependent = False

    def is_affected_by(self, reason):
        """Determine if TargetRid is affected by the update_type

        update_type is 'medium' or 'growth_rate'

        :param reason: reason of update, 'medium' or 'growth_rate'
        :type reason: str
        :return: Result of dependency check
        :rtype: bool
        """
        return ((reason == 'medium' and self.is_medium_dependent) or
                (reason == 'growth_rate' and self.is_growth_rate_dependent))


class InitialAssignments:

    def __init__(self, optim):
        """Instantiate Initial Assigments

        Handles SBML initial assignments
        Supports lambda function referencing SBML Function Definitions
        Initial assignment math gets expanded so resulting math can be evaluated using 'eval'
        Units, required in SBML for proper math formulation, get stripped

        :param optim: Optimization Instance with Cobra or Gurobi Model
        :type optim: Class Optimize
        """
        self.optim = optim
        m_dict = optim.m_dict

        self.local_env = {'np': np, 'growth_rate': 1.0}
        unit_ids = set(m_dict['unitDefs'].index) | {'dimensionless', 'second', 'hour', 'substance'}
        func_defs = self.get_function_defs(m_dict['funcDefs'], unit_ids)
        self.ia_functions = {symbol: IaFunction(symbol, row['math'])
                             for symbol, row in m_dict['initAssign'].iterrows()}

        # flag InitialAssignments which are growth and/or media dependent and expand function definition
        for symbol, iaf in self.ia_functions.items():
            iaf.is_medium_dependent = True if re.search(r'\b' + pf.M_, iaf.math) else False
            iaf.is_growth_rate_dependent = True if 'growth_rate' in iaf.math else False
            iaf.set_expanded_math(func_defs, unit_ids)
        self.target_rids = self.get_target_rids(m_dict['reactions'])

    @staticmethod
    def get_function_defs(df_func_defs, unit_ids):
        """Collect function definitons from the SBML model

        :param df_func_defs: Function definition data from SBML model
        :type df_func_defs: panda DataFrame
        :param unit_ids: unit ids used in the model, that may appear in math strings
        :type unit_ids: set of str
        :return: function definitions
        :rtype: dict
        """
        func_defs = {}
        for fid, row in df_func_defs.iterrows():

            # extract all, i.e. list of parameters and function definition, within brackets of lambda function
            m = re.search(r'\blambda\((.*)\)', sbmlxdf.misc.mathml2numpy(row['math']))
            params_fdef = m.group(1)

            # support function calls, which may contain commas, within the function definition
            idx = None
            parts = [kw.strip() for kw in params_fdef.split(',')]
            for idx, part in enumerate([kw.strip() for kw in parts]):
                if '(' in part:
                    break
            fdef_idx = min(idx, len(parts))
            params = parts[:fdef_idx]
            fdef = (params_fdef.split(',', fdef_idx)[fdef_idx]).strip()

            # remove any unit ids
            for uid in unit_ids:
                fdef = re.sub(r'\b' + uid + r'\b', '', fdef)
            func_defs[fid] = {'parameters': params, 'definition': fdef}
        return func_defs

    def get_target_rids(self, df_reactions):
        """Map reaction ids affected by initial assignment to related parameters

        matrix coefficients modified through initial assignment have
        'id' parameter set on respective species reference of reactants/products

        !! using original IDs

        :param df_reactions: reaction data of SBML model
        :type df_reactions: pandas Dataframe
        :return:
        """
        # identify target reactions of initial assignments
        ia_symbols = set(self.ia_functions)
        ia_flux_bounds = defaultdict(dict)
        ia_reactants = defaultdict(dict)
        ia_products = defaultdict(dict)
        for rid, row in df_reactions.iterrows():
            if row['fbcLowerFluxBound'] in ia_symbols:
                ia_flux_bounds[rid]['lb'] = row['fbcLowerFluxBound']
            if row['fbcUpperFluxBound'] in ia_symbols:
                ia_flux_bounds[rid]['ub'] = row['fbcUpperFluxBound']
            for sref_str in sbmlxdf.record_generator(row['reactants']):
                params = sbmlxdf.extract_params(sref_str)
                if 'id' in params and params['id'] in ia_symbols:
                    ia_reactants[rid].update({params['species']: params['id']})
            for sref_str in sbmlxdf.record_generator(row['products']):
                params = sbmlxdf.extract_params(sref_str)
                if 'id' in params and params['id'] in ia_symbols:
                    ia_products[rid].update({params['species']: params['id']})

        # create target rids
        rids = set(ia_flux_bounds) | set(ia_reactants) | set(ia_products)
        target_rids = {rid: IaTargetRid(rid) for rid in rids}
        for rid, data in ia_flux_bounds.items():
            target_rids[rid].flux_bounds = data
        for rid, data in ia_reactants.items():
            target_rids[rid].reactants = data
        for rid, data in ia_products.items():
            target_rids[rid].products = data

        # set medium / growth rate dependency on reaction targeted by initial assignments
        for trid, tr in target_rids.items():
            symbols = set(tr.flux_bounds.values()) | set(tr.reactants.values()) | set(tr.products.values())
            for symbol in symbols:
                if self.ia_functions[symbol].is_medium_dependent:
                    target_rids[trid].is_medium_dependent = True
                if self.ia_functions[symbol].is_growth_rate_dependent:
                    target_rids[trid].is_growth_rate_dependent = True
        return target_rids

    def set_medium(self, medium_concs):
        """Set medium dependent parameters in RBA model.

        Medium dependent model parameters have sids as variable names in
        their initial assignment math string.

        :param medium_concs: species concentrations in mmol/l
        :type medium_concs: dict (key: sid, val: float)
        """
        for sid, conc in medium_concs.items():
            self.local_env[sid] = conc
        self.update_model_parameters('medium')

    def set_growth_rate(self, growth_rate):
        """Set growth rate dependent parameters in RBA model.

        Growth rate dependent model parameters have 'growth_rate' as variable name in
        their initial assignment math string.

        Only update of problem parameters when values have changed

        :param growth_rate: selected growth rate in h-1
        :type growth_rate: float
        """
        self.local_env['growth_rate'] = growth_rate
        self.update_model_parameters('growth_rate')

    def update_model_parameters(self, reason):
        if self.optim.is_gpm:
            self.gp_update_model_parameters(reason)
        else:
            self.cp_update_model_parameters(reason)

    def gp_update_model_parameters(self, reason):
        """GurobiPy model: modify flux bounds, coefficients based on initial assignments.

        :param reason: reason for update ('medium' or 'growth_rate')
        :type reason: str
        """
        gpm = self.optim.gpm
        gpm.update()

        for rid, tr in self.target_rids.items():
            if tr.is_affected_by(reason):
                var = gpm.getVarByName(rid)

                # configure variable bounds that depend on growth rate and medium
                if len(tr.flux_bounds) > 0:
                    if 'lb' in tr.flux_bounds:
                        new_lb = self.ia_functions[tr.flux_bounds['lb']].get_value(self.local_env)
                        if new_lb != var.lb:
                            var.lb = new_lb
                    if 'ub' in tr.flux_bounds:
                        new_ub = self.ia_functions[tr.flux_bounds['ub']].get_value(self.local_env)
                        if new_ub != var.ub:
                            var.ub = new_ub

                for sid, symbol_id in tr.reactants.items():
                    new_val = -self.ia_functions[symbol_id].get_value(self.local_env)
                    constr = gpm.getConstrByName(sid)
                    if new_val != gpm.getCoeff(constr, var):
                        gpm.chgCoeff(constr, var, new_val)
                for sid, symbol_id in tr.products.items():
                    new_val = self.ia_functions[symbol_id].get_value(self.local_env)
                    constr = gpm.getConstrByName(sid)
                    if new_val != gpm.getCoeff(constr, var):
                        gpm.chgCoeff(constr, var, new_val)

    def cp_update_model_parameters(self, reason):
        """CobraPy model: modify flux bounds, coefficients based on initial assignments.

        :param reason: reason for update ('medium' or 'growth_rate')
        :type reason: str
        """
        model = self.optim.model
        for rid, tr in self.target_rids.items():
            if tr.is_affected_by(reason):
                ridx = re.sub(f'^{pf.R_}', '', rid)
                rxn = model.reactions.get_by_id(ridx)

                # configure variable bounds that deped on growth rate and medium
                if len(tr.flux_bounds) > 0:
                    if 'lb' in tr.flux_bounds and 'ub' in tr.flux_bounds:
                        new_lb = self.ia_functions[tr.flux_bounds['lb']].get_value(self.local_env)
                        new_ub = self.ia_functions[tr.flux_bounds['ub']].get_value(self.local_env)
                        rxn.bounds = (new_lb, new_ub)
                    elif 'lb' in tr.flux_bounds:
                        new_lb = self.ia_functions[tr.flux_bounds['lb']].get_value(self.local_env)
                        rxn.lower_bound = new_lb
                    elif 'ub' in tr.flux_bounds:
                        new_ub = self.ia_functions[tr.flux_bounds['ub']].get_value(self.local_env)
                        rxn.lower_bound = new_ub

                # configure stoichiometric coefficients that depend on growth rate or medium
                old_mids = {met.id for met in rxn.metabolites}
                new_coefs = {}
                if len(tr.reactants) > 0:
                    new_coefs = {re.sub(f'^{pf.M_}', '', sid): -self.ia_functions[symbol_id].get_value(self.local_env)
                                 for sid, symbol_id in tr.reactants.items()}
                if len(tr.products) > 0:
                    new_coefs |= {re.sub(f'^{pf.M_}', '', sid): self.ia_functions[symbol_id].get_value(self.local_env)
                                  for sid, symbol_id in tr.products.items()}

                # in case of newly added metabolites (i.e. currently not existing in the reaction)
                #  use combine=True in rxn.add_metabolites(), otherwise use combine=False to improve accuracy.
                # Note: CobraPy 0.27.0 does not support adding new metabolites with combine=False wrt contexts
                #   mets_to_reset = { key: old_coefficients[model.metabolites.get_by_any(key)[0]]
                #                     for key in metabolites_to_add.keys()}
                #   will result in a key_error for metabolites that were not part of the reaction
                new_srefs = {sidx: coef for sidx, coef in new_coefs.items() if sidx not in old_mids}
                modified_srefs = {sidx: coef for sidx, coef in new_coefs.items() if sidx in old_mids}
                if len(new_srefs) > 0:
                    rxn.add_metabolites(new_srefs, combine=True)
                if len(modified_srefs) > 0:
                    rxn.add_metabolites(modified_srefs, combine=False)
